/*
 * Copyright (c) 2007 Anthony Mallet                  --  Tue Jan 23 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdlib.h>
#include <portLib.h>

#include <viamlib.h>

#include "server/viamHeader.h"
#include "shared.h"


/* --- viamBusPrintCntrl --------------------------------------------- */

/** \brief Control codel of CONTROL request BusPrint
 *
 * Report: OK, S_viam_FWBUS
 *
 * \return OK or ERROR
 */

STATUS
viamBusPrintCntrl(int *report)
{
  if (viam_hardware_list(vh))
    VIAM_RETURN_WITHERRNO(report, ERROR);

  return OK;
}


/* --- viamBankCreateCntrl ------------------------------------------- */

/** \brief Control codel of CONTROL request BankCreate
 *
 * Add a bank to the configuration
 *
 * \return OK or ERROR
 */

STATUS
viamBankCreateCntrl(ViamBankCreate *bank, int *report)
{
  viam_bank_t b;
  viam_bankdata_t *d;

  VIAM_RETURN_IFRUNNING(report, ERROR);

  /* create bank in viam libs */
  b = viam_bank_create(vh, bank->name.id);
  if (!b) VIAM_RETURN_WITHERRNO(report, ERROR);

  /* initialize local data */
  d = malloc(sizeof(*d));
  if (!d) {
    viam_bank_delete(vh, b);
    *report = S_viam_MEMORY;
    return ERROR;
  }

  d->buffers = bank->buffering;
  d->tags = bank->tags;
  d->pdata = NULL;
  d->idata = NULL;
  d->nidata = 0;
  viam_bank_setusrdata(b, d);

  return OK;
}


/* --- viamCameraCreateCntrl ----------------------------------------- */

/** \brief Control codel of CONTROL request CameraCreate
 *
 * Add a camera to the configuration
 *
 * \return OK or ERROR
 */

STATUS
viamCameraCreateCntrl(ViamCameraCreate *c, int *report)
{
  VIAM_RETURN_IFRUNNING(report, ERROR);

  if (!viam_camera_create(vh, c->uid, c->name.id))
    VIAM_RETURN_WITHERRNO(report, ERROR);

  return OK;
}


/*------------------------------------------------------------------------
 * viamBankAddCameraCntrl  -  control codel of CONTROL request BankAddCamera
 *
 * Description:
 * Report: OK
 *                 S_viam_BAD_ID
 *                 S_viam_MODULE_RUNNING
 *
 * Returns:    OK or ERROR
 */

STATUS
viamBankAddCameraCntrl(ViamBankAddCamera *bank_add_camera, int *report)
{
  viam_bank_t bank;
  viam_camera_t camera;

  VIAM_RETURN_IFRUNNING(report, ERROR);
  bank = viam_bank_getbyname(vh, bank_add_camera->bank.id);
  if (!bank)
    VIAM_RETURN_WITHERRNO(report, ERROR);

  camera = viam_camera_getbyname(vh, bank_add_camera->camera.id);
  if (!camera)
    VIAM_RETURN_WITHERRNO(report, ERROR);

  if (viam_bank_cameraadd(vh, bank, camera, bank_add_camera->name.id))
    VIAM_RETURN_WITHERRNO(report, ERROR);

  return OK;
}


/* --- viamCameraPushGeoFilterCntrl ---------------------------------- */

/** \brief control codel of CONTROL request CameraPushGeoFilter
 *
 * Report:	   OK
 *                 S_viam_NO_SUCH_CAMERA
 *                 S_viam_CONFIGURED
 *
 * \return OK or ERROR
 */

STATUS
viamCameraPushGeoFilterCntrl(ViamGeoFilter *geofilter, int *report)
{
  viam_image_t image;
  viam_filter_t filter;
  int s;

  image = viam_image_getbyname(vh, geofilter->image.id);
  if (!image) VIAM_RETURN_WITHERRNO(report, ERROR);

  switch(geofilter->type) {
    case VIAM_FILTER_SUBSAMPLE:
      filter = viam_filter_subsample_create(
	vh, geofilter->left, geofilter->top, geofilter->right,
	geofilter->bottom, geofilter->toWidth, geofilter->toHeight);
      break;

#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_DISTORT:
      filter = viam_filter_distort_create(
	vh, geofilter->left, geofilter->top, geofilter->right,
	geofilter->bottom, geofilter->toWidth, geofilter->toHeight);
      break;

    case VIAM_FILTER_RECTIFY:
      filter = viam_filter_rectify_create(
	vh, geofilter->left, geofilter->top, geofilter->right,
	geofilter->bottom, geofilter->toWidth, geofilter->toHeight);
      break;
#endif /* VIAM_OPENCV_SUPPORT */

    default:
      *report = S_viam_NOT_IMPLEMENTED;
      return ERROR;
  }
  if (!filter) VIAM_RETURN_WITHERRNO(report, ERROR);

  s = viam_filter_push(vh, image, filter, geofilter->method, geofilter->automode);
  if (s) VIAM_RETURN_WITHERRNO(report, ERROR);

  return OK;
}


/* --- viamCameraPushColorFilterCntrl -------------------------------- */

/** \brief Control codel of CONTROL request CameraPushColorFilter
 *
 * Report: OK
 *                 S_viam_NO_SUCH_CAMERA
 *                 S_viam_CONFIGURED
 *                 S_viam_INVALID_VALUE
 *                 S_viam_NOT_IMPLEMENTED
 *
 * \returns OK or ERROR
 */

STATUS
viamCameraPushColorFilterCntrl(ViamColorFilter *colorfilter, int *report)
{
  viam_image_t image;
  viam_filter_t filter;
  int s;

  image = viam_image_getbyname(vh, colorfilter->image.id);
  if (!image) VIAM_RETURN_WITHERRNO(report, ERROR);

  switch(colorfilter->type) {
#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_WHITE_BALANCE:
      filter = viam_filter_whitebalance_create(
	vh,
	((int)colorfilter->parameter)/100.,
	colorfilter->parameter - (int)colorfilter->parameter);
      break;

    case VIAM_FILTER_BAYER_BGGR:
    case VIAM_FILTER_BAYER_GBRG:
    case VIAM_FILTER_BAYER_RGGB:
    case VIAM_FILTER_BAYER_GRBG:
      filter = viam_filter_bayer_create(vh, colorfilter->type);
      break;
#endif /* VIAM_OPENCV_SUPPORT */

    default:
      *report = S_viam_NOT_IMPLEMENTED;
      return ERROR;
  }
  if (!filter) VIAM_RETURN_WITHERRNO(report, ERROR);

  s = viam_filter_push(vh, image,
		       filter, colorfilter->method, colorfilter->automode);
  if (s) VIAM_RETURN_WITHERRNO(report, ERROR);

  return OK;
}


/* --- viamCameraSetHWModeMain --------------------------------------- */

/** \brief control codel of CONTROL request CameraSetHWMode
 *
 * Report: OK
 *                 S_viam_NO_SUCH_CAMERA
 *                 S_viam_CONFIGURED
 *
 * \returns OK or ERROR
 */

STATUS
viamCameraSetHWModeMain(ViamHWMode *hwmode, int *report)
{
  viam_camera_t camera = viam_camera_getbyname(vh, hwmode->camera.id);
  if (!camera) VIAM_RETURN_WITHERRNO(report, ERROR);

  if (viam_camera_sethwmode(vh, camera, &hwmode->mode))
    VIAM_RETURN_WITHERRNO(report, ERROR);

  return OK;
}




/* --- CalibrationIO ------------------------------------------------- */

/** \brief control codel of CONTROL request CalibrationIO
 *
 * Reports:      OK
 *              S_viam_NO_SUCH_BANK
 *              S_viam_NOT_CONFIGURED
 *              S_viam_CALIBRATING
 *              S_viam_NOT_CALIBRATED
 *              S_viam_FILE_IO
 * \return EXEC END ETHER FAIL ZOMBIE
 */

STATUS
viamCalibrationIOMain(ViamCalibrationIO *calibrationio, int *report)
{
  viam_bank_t bank;
  int s;

  bank = viam_bank_getbyname(vh, calibrationio->bank.id);
  if (!bank) VIAM_RETURN_WITHERRNO(report, ERROR);

  switch(calibrationio->op) {
    case VIAM_LOAD:
      s = viam_calibration_load(vh, bank, calibrationio->file.id);
      break;

    case VIAM_SAVE:
      s = viam_calibration_save(vh, bank, calibrationio->file.id);
      break;
  }
  if (s) VIAM_RETURN_WITHERRNO(report, ERROR);
  return ETHER;
}


/* --- viamDisplayCntrl ---------------------------------------------- */

/** \brief control codel of CONTROL request Display
 *
 * Report: OK, S_viam_OPENCV
 *
 * \return OK or ERROR
 */

STATUS
viamDisplayCntrl(ViamDisplay *display, int *report)
{
#ifdef VIAM_DISPLAY
  viam_bank_t bank;
  viam_image_t image;
  viam_imagedata_t *idata;
  int i, s;

  /* find bank */
  bank = viam_bank_getbyname(vh, display->bank.id);
  if (!bank) VIAM_RETURN_WITHERRNO(report, ERROR);
  if (!viam_bank_isconfigured(bank)) {
    *report = S_viam_NOT_CONFIGURED;
    return ERROR;
  }

  /* find image */
  for(i=0; i<viam_bank_nimages(bank); i++) {
    image = viam_bank_image(vh, bank, i);
    if (!strcmp(viam_image_name(image), display->image.id))
      break;
    image = NULL;
  }
  if (!image) { *report = S_viam_NO_SUCH_IMAGE; return ERROR; }
  idata = viamIData(bank, i);
  if (!idata) { *report = S_viam_MEMORY; return ERROR; }

  /* create/destroy window */
  if (idata->window) {
    cvDestroyWindow(idata->window);
    free(idata->window);
    idata->window = NULL;
  }

  if (display->enable == VIAM_ON) {
    idata->window = malloc(
      strlen(display->bank.id) + strlen(" - ") + strlen(display->image.id)
      + 1);
    if (!idata->window) { *report = S_viam_NO_SUCH_IMAGE; return ERROR; }
    sprintf(idata->window, "%s - %s", display->bank.id, display->image.id);

    cvNamedWindow(idata->window,
		  (display->width>0 || display->height>0)?0:CV_WINDOW_AUTOSIZE);
    if (cvGetErrStatus()) { *report = S_viam_OPENCV; return ERROR; }

    if (display->width > 0 || display->height > 0) {
      if (display->height == 0 || display->width == 0) {
	viam_geometry_t geom;

	s = viam_image_geometry(vh, image, &geom);
	if (s) VIAM_RETURN_WITHERRNO(report, ERROR);

	if (display->width == 0)
	  display->width = geom.width * display->height / geom.height;
	if (display->height == 0)
	  display->height = geom.height * display->width / geom.width;
      }

      cvResizeWindow(idata->window, display->width, display->height);
    }
  }

  /* store misc parameters */
  idata->vtags = display->vtag;

  return OK;
#else
  *report = S_viam_FEATURE_DISABLED;
  return ERROR;
#endif /* VIAM_DISPLAY */
}


/* --- viamSaveCntrl ------------------------------------------------- */

/** \brief control codel of CONTROL request Save
 *
 * \return OK or ERROR
 */

STATUS
viamSaveCntrl(ViamSave *save, int *report)
{
  viam_bank_t bank;
  viam_image_t image;
  viam_imagedata_t *idata;
  int i;

  /* find bank */
  bank = viam_bank_getbyname(vh, save->bank.id);
  if (!bank) VIAM_RETURN_WITHERRNO(report, ERROR);
  if (!viam_bank_isconfigured(bank)) {
    *report = S_viam_NOT_CONFIGURED;
    return ERROR;
  }

  /* find image */
  for(i=0; i<viam_bank_nimages(bank); i++) {
    image = viam_bank_image(vh, bank, i);
    if (!strcmp(viam_image_name(image), save->image.id))
      break;
    image = NULL;
  }
  if (!image) { *report = S_viam_NO_SUCH_IMAGE; return ERROR; }
  idata = viamIData(bank, i);
  if (!idata) { *report = S_viam_MEMORY; return ERROR; }

  /* create/destroy window */
  switch (save->enable) {
    case VIAM_ON:
      if (!idata->file) {
	idata->file = strdup(save->pattern.id);
	idata->seq = 0;
      }
      break;

    case VIAM_OFF:
      if (idata->file) {
	free(idata->file);
	idata->file = NULL;
      }
      break;
  }
  return OK;
}
